/*
 * miscellaneous string utility functions
 * 
 * Copyright (c) 2001, 2002 xdvik project team
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL PAUL VOJTA OR ANY OTHER AUTHOR OF THIS SOFTWARE BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "xdvi-config.h"
#include "xdvi.h"
#include "string-utils.h"
#include "util.h"


/*------------------------------------------------------------
 *  str_is_prefix
 *
 *  Purpose:
 *	Check if <str1> is a prefix of <str2> or vice versa.
 *      Returns True if it is, False if it isn't.
 *------------------------------------------------------------*/

Boolean str_is_prefix(const char *str1, const char *str2)
{
    int i;

    for (i = 0; *str1 != '\0' && *str2 != '\0'; i++) {
	if (*str1++ != *str2++) {
	    return False;
	}
    }
    return True;
}



/*------------------------------------------------------------
 *  str_is_suffix
 *
 *  Purpose:
 *	Check if <str1> is a suffix of <str2> or vice versa.
 *      Returns True if it is, False if it isn't.
 *------------------------------------------------------------*/

Boolean str_is_suffix(const char *str1, const char *str2)
{
    int len1 = strlen(str1);
    int len2 = strlen(str2);

    while (len1 > len2) {
	str1++;
	len1--;
    }
    while (len2 > len1) {
	str2++;
	len2--;
    }
    if (strcmp(str1, str2) == 0) {
	return True;
    }
    else {
	return False;
    }
}

/*------------------------------------------------------------
 *  length_of_int, length_of_ulong
 *
 *  Purpose:
 *	Return length of printed representation of integer or long <n>,
 *	including the minus sign for negative digits.
 *------------------------------------------------------------*/

int
length_of_int(int n)
{
    if (n == 0)
	return 1;

    if (n < 0) {
	unsigned long ul = n * -1;
	return length_of_ulong(ul) + 1;
    }
    return length_of_ulong((unsigned long)n);
}


int
length_of_ulong(unsigned long n)
{
    int ret = 0;

    if (n == 0)
	return 1;

    while (n >= 1) {
	n /= 10;
	ret++;
    }
    return ret;
}

/*
  expand filename to include  full path name;
  returns result in a freshly allocated string.
*/
char *
expand_filename(const char *filename)
{
    char *path_name = NULL;
    size_t path_name_len = 512;
    size_t len = strlen(filename) + 1;

    /* skip over `./' in relative paths */
    if (filename[0] == '.' && filename[1] == '/') {
	filename += 2;
    }

    /* expand to full path if needed */
    if (filename[0] != '/') {
	for (;;) {
	    char *tmp;
	    path_name = xrealloc(path_name, path_name_len);
	    if ((tmp = getcwd(path_name, path_name_len)) == NULL && errno == ERANGE) {
		path_name_len *= 2;
	    }
	    else {
		path_name = tmp;
		break;
	    }
	}
	len += strlen(path_name) + 1;	/* 1 for '/' */
	path_name = xrealloc(path_name, len);
	strcat(path_name, "/");
	strcat(path_name, filename);
    }

    return path_name ? path_name : xstrdup(filename);
}

/* expand filename to include `.dvi' extension and full path name;
   returns malloc()ed string (caller is responsible for free()ing).
*/
char *
normalize_and_expand_filename(const char *filename)
{
    char *expanded_filename = NULL;
    size_t len;
    char *ret;
    
    /* skip over `file:' prefix if present */
    if (memcmp(filename, "file:", 5) == 0) {
	filename += 5;
    }

    len = strlen(filename) + 5; /* 5 in case we need to add `.dvi\0' */

    expanded_filename = xmalloc(len);

    strcpy(expanded_filename, filename);

    /* append ".dvi" extension if not present */
    if (!str_is_suffix(expanded_filename, ".dvi")) {
	strcat(expanded_filename, ".dvi");
    }

    ret = expand_filename(expanded_filename);
    free(expanded_filename);
    return ret;
}


char *
format_arg(const char *fmt, const char *arg, int *match)
{
    char *tmp = xmalloc(strlen(fmt) + strlen(arg) + 1);
    if (strchr(fmt, '%') != NULL) {
	sprintf(tmp, fmt, arg);
	*match = 1;
    }
    else {
	strcpy(tmp, fmt);
    }
    return tmp;
}

/* escape single `%' characters in arg and return newly allocated string */
char *
escape_format_arg(const char *arg)
{
    char *ret, *ptr;
    ASSERT(arg != NULL);
    ret = xmalloc(strlen(arg) * 2 + 1); /* more than enuff */
    
    ptr = ret;
    while (*arg != '\0') {
	/* need to escape? */
	if (*arg == '%' && (ptr == ret
			    || (ptr > ret && *(arg - 1) != '%'))) {
	    *ptr++ = '%';
	}
	*ptr++ = *arg++;
    }
    *ptr = '\0';

    /* trim return buffer */
    return xrealloc(ret, strlen(ret) + 1);
}

char *
unquote_arg(char *fmt, const char *arg, int *match, int *len)
{
    char *ptr;
    char c;
    
    c = fmt[0];
    fmt++;
    if ((ptr = strchr(fmt, c)) != NULL) {
	*ptr++ = '\0';
	while (*ptr == ' ' || *ptr == '\t') {
	    ptr++;
	}
	*len = ptr - fmt;
	return format_arg(fmt, arg, match);
    }
    else {
	*len = strlen(fmt);
	XDVI_WARNING((stderr, "Ignoring lonesome quote in string %s", fmt - 1));
	return format_arg(fmt, arg, match);
    }
}

/* chop source into space-separated chunks and save these to newly allocated return list */
char **
get_space_separated_list(const char *source)
{
    char **chunks = NULL;
    const char *b_ptr, *e_ptr;
    size_t alloc_len = 0, i = 0, len = 0, chunk_len;
    b_ptr = source;
    while (*b_ptr != '\0' && (len = strcspn(b_ptr, " \t")) > 0) {
	e_ptr = b_ptr + len;
	chunk_len = e_ptr - b_ptr;
	while (i + 1 >= alloc_len) {
	    alloc_len++;
	}
	chunks = xrealloc(chunks, alloc_len * sizeof *chunks);
	chunks[i] = xmalloc(chunk_len + 1);
	memcpy(chunks[i], b_ptr, chunk_len);
	chunks[i][chunk_len] = '\0';
	i++;

	e_ptr += strspn(e_ptr, " \t");
	if (*e_ptr == '\0')
	    break;
	b_ptr = e_ptr;
    }
    chunks[i] = NULL;
    return chunks;
}

const char *
find_format_str(const char *input, const char *fmt)
{
    const char *ptr = input;
    while ((ptr = strstr(ptr, fmt)) != NULL) {
	if (ptr > input && *(ptr - 1) == '\\') {
	    ptr++;
	    continue;
	}
	else
	    return ptr;
    }
    return NULL;
}
